@@ -60,6 +60,8 @@ module Language.Haskell.TH.Datatype
   , DatatypeVariant(..)
   , ConstructorVariant(..)
   , FieldStrictness(..)
+  , Unpackedness(..)
+  , Strictness(..)
 
   -- * Normalization functions
   , reifyDatatype
@@ -86,6 +88,9 @@ module Language.Haskell.TH.Datatype
 
   -- * Backward compatible data definitions
   , dataDCompat
+  , newtypeDCompat
+  , tySynInstDCompat
+  , pragLineDCompat
   , arrowKCompat
 
   -- * Strictness annotations
@@ -106,6 +111,7 @@ module Language.Haskell.TH.Datatype
   -- * Convenience functions
   , unifyTypes
   , tvName
+  , tvKind
   , datatypeType
   ) where
 
@@ -539,7 +545,7 @@ repairVarKindsWith = zipWith stealKindForType
 
 -- If a VarT is missing an explicit kind signature, steal it from a TyVarBndr.
 stealKindForType :: TyVarBndr -> Type -> Type
-stealKindForType tvb t@VarT{} = SigT t (tvbKind tvb)
+stealKindForType tvb t@VarT{} = SigT t (tvKind tvb)
 stealKindForType _   t        = t
 
 -- | Normalize 'Dec' for a newtype or datatype into a 'DatatypeInfo'.
@@ -611,9 +617,9 @@ bndrParams = map $ \bndr ->
     PlainTV  t   -> VarT t
 
 -- | Extract the kind from a 'TyVarBndr'. Assumes 'PlainTV' has kind @*@.
-tvbKind :: TyVarBndr -> Kind
-tvbKind (PlainTV  _)   = starK
-tvbKind (KindedTV _ k) = k
+tvKind :: TyVarBndr -> Kind
+tvKind (PlainTV  _)   = starK
+tvKind (KindedTV _ k) = k
 
 -- | Remove the outermost 'SigT'.
 stripSigT :: Type -> Type
@@ -806,17 +812,18 @@ normalizeConFor reifiedDec typename params variant = fmap (map giveTyVarBndrsSta
           mightHaveBeenEtaReduced ts =
             case unsnoc ts of
               Nothing -> False
-              Just (initTs,lastT) ->
+              Just (initTs :|- lastT) ->
                 case varTName lastT of
                   Nothing -> False
                   Just n  -> not (n `elem` freeVariables initTs)
 
-          -- If the list is empty returns 'Nothing', otherwise returns the 'init' and the 'last'.
-          unsnoc :: [a] -> Maybe ([a], a)
+          -- If the list is empty returns 'Nothing', otherwise returns the
+          -- 'init' and the 'last'.
+          unsnoc :: [a] -> Maybe (NonEmptySnoc a)
           unsnoc [] = Nothing
-          unsnoc [x] = Just ([], x)
-          unsnoc (x:xs) = Just (x:a, b)
-              where Just (a,b) = unsnoc xs
+          unsnoc (x:xs) = case unsnoc xs of
+            Just (a :|- b) -> Just ((x:a) :|- b)
+            Nothing        -> Just ([]    :|- x)
 
           -- If a Type is a VarT, find Just its Name. Otherwise, return Nothing.
           varTName :: Type -> Maybe Name
@@ -878,16 +885,36 @@ normalizeGadtC ::
   Q [ConstructorInfo]
 normalizeGadtC typename params tyvars context names innerType
                fields stricts getVariant =
-  do innerType' <- resolveTypeSynonyms innerType
+  do -- Due to GHC Trac #13885, it's possible that the type variables bound by
+     -- a GADT constructor will shadow those that are bound by the data type.
+     -- This function assumes this isn't the case in certain parts (e.g., when
+     -- mergeArguments is invoked), so we do an alpha-renaming of the
+     -- constructor-bound variables before proceeding. See #36 for an example
+     -- of what can go wrong if this isn't done.
+     let conBoundNames =
+           concatMap (\tvb -> tvName tvb:freeVariables (tvKind tvb)) tyvars
+     conSubst <- T.sequence $ Map.fromList [ (n, newName (nameBase n))
+                                           | n <- conBoundNames ]
+     let conSubst'     = fmap VarT conSubst
+         renamedTyvars =
+           map (\tvb -> case tvb of
+                          PlainTV n    -> PlainTV  (conSubst Map.! n)
+                          KindedTV n k -> KindedTV (conSubst Map.! n)
+                                                   (applySubstitution conSubst' k)) tyvars
+         renamedContext   = applySubstitution conSubst' context
+         renamedInnerType = applySubstitution conSubst' innerType
+         renamedFields    = applySubstitution conSubst' fields
+
+     innerType' <- resolveTypeSynonyms renamedInnerType
      case decomposeType innerType' of
        ConT innerTyCon :| ts | typename == innerTyCon ->
 
          let (substName, context1) = mergeArguments params ts
              subst   = VarT <$> substName
-             tyvars' = [ tv | tv <- tyvars, Map.notMember (tvName tv) subst ]
+             tyvars' = [ tv | tv <- renamedTyvars, Map.notMember (tvName tv) subst ]
 
-             context2 = applySubstitution subst (context1 ++ context)
-             fields'  = applySubstitution subst fields
+             context2 = applySubstitution subst (context1 ++ renamedContext)
+             fields'  = applySubstitution subst renamedFields
          in sequence [ ConstructorInfo name tyvars' context2
                                        fields' stricts <$> variantQ
                      | name <- names
@@ -1156,9 +1183,14 @@ instance TypeSubstitution Type where
   applySubstitution subst = go
     where
       go (ForallT tvs context t) =
-        let subst' = foldl' (flip Map.delete) subst (map tvName tvs) in
-        ForallT tvs (applySubstitution subst' context)
-                    (applySubstitution subst' t)
+        let subst' = foldl' (flip Map.delete) subst (map tvName tvs)
+
+            mapTvbKind :: (Kind -> Kind) -> TyVarBndr -> TyVarBndr
+            mapTvbKind f (PlainTV n)    = PlainTV n
+            mapTvbKind f (KindedTV n k) = KindedTV n (f k) in
+        ForallT (map (mapTvbKind (applySubstitution subst')) tvs)
+                (applySubstitution subst' context)
+                (applySubstitution subst' t)
       go (AppT f x)      = AppT (go f) (go x)
       go (SigT t k)      = SigT (go t) (applySubstitution subst k) -- k could be Kind
       go (VarT v)        = Map.findWithDefault (VarT v) v subst
@@ -1172,7 +1204,9 @@ instance TypeSubstitution Type where
   freeVariables t =
     case t of
       ForallT tvs context t' ->
-          (freeVariables context `union` freeVariables t')
+          (concatMap (freeVariables . tvKind) tvs
+              `union` freeVariables context
+              `union` freeVariables t')
           \\ map tvName tvs
       AppT f x      -> freeVariables f `union` freeVariables x
       SigT t' k     -> freeVariables t' `union` freeVariables k
@@ -1397,6 +1431,50 @@ dataDCompat c n ts cs ds =
 dataDCompat = dataD
 #endif
 
+-- | Backward compatible version of 'newtypeD'
+newtypeDCompat ::
+  CxtQ        {- ^ context                 -} ->
+  Name        {- ^ type constructor        -} ->
+  [TyVarBndr] {- ^ type parameters         -} ->
+  ConQ        {- ^ constructor definition  -} ->
+  [Name]      {- ^ derived class names     -} ->
+  DecQ
+#if MIN_VERSION_template_haskell(2,12,0)
+newtypeDCompat c n ts cs ds =
+  newtypeD c n ts Nothing cs
+    (if null ds then [] else [derivClause Nothing (map conT ds)])
+#elif MIN_VERSION_template_haskell(2,11,0)
+newtypeDCompat c n ts cs ds =
+  newtypeD c n ts Nothing cs
+    (return (map ConT ds))
+#else
+newtypeDCompat = newtypeD
+#endif
+
+-- | Backward compatible version of 'tySynInstD'
+tySynInstDCompat ::
+  Name    {- ^ type family name    -} ->
+  [TypeQ] {- ^ instance parameters -} ->
+  TypeQ   {- ^ instance result     -} ->
+  DecQ
+#if MIN_VERSION_template_haskell(2,9,0)
+tySynInstDCompat n ps r = TySynInstD n <$> (TySynEqn <$> sequence ps <*> r)
+#else
+tySynInstDCompat = tySynInstD
+#endif
+
+-- | Backward compatible version of 'pragLineD'. Returns
+-- 'Nothing' if line pragmas are not suported.
+pragLineDCompat ::
+  Int     {- ^ line number -} ->
+  String  {- ^ file name   -} ->
+  Maybe DecQ
+#if MIN_VERSION_template_haskell(2,10,0)
+pragLineDCompat ln fn = Just (pragLineD ln fn)
+#else
+pragLineDCompat _  _  = Nothing
+#endif
+
 arrowKCompat :: Kind -> Kind -> Kind
 #if MIN_VERSION_template_haskell(2,8,0)
 arrowKCompat x y = arrowK `appK` x `appK` y
