@@ -23,15 +23,24 @@ the versions of GHC supported by this package.
 module Main (main) where
 
 #if __GLASGOW_HASKELL__ >= 704
-import Control.Monad (zipWithM_)
+import           Control.Monad (zipWithM_)
 #endif
 
-import Language.Haskell.TH
-import Language.Haskell.TH.Datatype
-import Language.Haskell.TH.Lib (starK)
+#if MIN_VERSION_template_haskell(2,8,0)
+import           Control.Monad (unless)
+import qualified Data.Map as Map
+#endif
 
-import Harness
-import Types
+#if MIN_VERSION_base(4,7,0)
+import           Data.Type.Equality ((:~:)(..))
+#endif
+
+import           Language.Haskell.TH
+import           Language.Haskell.TH.Datatype
+import           Language.Haskell.TH.Lib (starK)
+
+import           Harness
+import           Types
 
 -- | Test entry point. Tests will pass or fail at compile time.
 main :: IO ()
@@ -62,6 +71,12 @@ main =
 #endif
      reifyDatatypeWithConNameTest
      reifyConstructorTest
+#if MIN_VERSION_base(4,7,0)
+     importedEqualityTest
+#endif
+#if MIN_VERSION_template_haskell(2,8,0)
+     kindSubstTest
+#endif
 
 adt1Test :: IO ()
 adt1Test =
@@ -521,7 +536,6 @@ resolvePredSynonymsTest =
 reifyDatatypeWithConNameTest :: IO ()
 reifyDatatypeWithConNameTest =
   $(do info <- reifyDatatype 'Just
-       let a = VarT (mkName "a")
        validateDI info
          DatatypeInfo
           { datatypeContext = []
@@ -546,3 +560,49 @@ reifyConstructorTest :: IO ()
 reifyConstructorTest =
   $(do info <- reifyConstructor 'Just
        validateCI info justCI)
+
+#if MIN_VERSION_base(4,7,0)
+importedEqualityTest :: IO ()
+importedEqualityTest =
+  $(do info <- reifyDatatype ''(:~:)
+       let [a,b] = map (VarT . mkName) ["a","b"]
+           k     = mkName "k"
+           kKind = varKCompat k
+       validateDI info
+         DatatypeInfo
+           { datatypeContext = []
+           , datatypeName    = ''(:~:)
+           , datatypeVars    = [SigT a kKind, SigT b kKind]
+           , datatypeVariant = Datatype
+           , datatypeCons    =
+               [ ConstructorInfo
+                   { constructorName       = 'Refl
+                   , constructorVars       = [KindedTV k starK]
+                     -- This shouldn't happen, ideally. See #37.
+
+                   , constructorContext    = [equalPred a b]
+                   , constructorFields     = []
+                   , constructorStrictness = []
+                   , constructorVariant    = NormalConstructor } ]
+           }
+   )
+#endif
+
+#if MIN_VERSION_template_haskell(2,8,0)
+kindSubstTest :: IO ()
+kindSubstTest =
+  $(do k1 <- newName "k1"
+       k2 <- newName "k2"
+       a  <- newName "a"
+       let ty = ForallT [KindedTV a (VarT k1)] [] (VarT a)
+           substTy = applySubstitution (Map.singleton k1 (VarT k2)) ty
+
+           checkFreeVars :: Type -> [Name] -> Q ()
+           checkFreeVars t freeVars =
+             unless (freeVariables t == freeVars) $
+               fail $ "free variables of " ++ show t ++ " should be " ++ show freeVars
+
+       checkFreeVars ty      [k1]
+       checkFreeVars substTy [k2]
+       [| return () |])
+#endif
